/*
 * This tests an app flow on the cloudron. User installs
 * an app, tries to configure and then uninstall it.
 */

'use strict';

var AppStore = require('../appstore.js'),
    assert = require('assert'),
    Cloudron = require('../cloudron.js'),
    common = require('../common.js'),
    net = require('net');

require('colors');

process.env.NODE_TLS_REJECT_UNAUTHORIZED = '0';
var BOX_VERSION = process.env.BOX_VERSION;

describe('Application flow test', function () {
    this.timeout(0);

    var appStore = new AppStore('https://api.staging.cloudron.io');

    var owner = common.getOwner();
    var admin = common.getAdmin();
    var cloudron, appId, box;

    it('can login to the store', function () {
        var accessToken = appStore.getAccessToken(owner);
        appStore.setCredentials(owner.password, accessToken);

        var adminAccessToken = appStore.getAccessToken(admin);
        appStore.setAdminCredentials(admin.password, adminAccessToken);
    });

    it('can create a cloudron', function () {
        box = appStore.createCloudron({
            domain: common.cloudronDomain(__filename),
            region: 'sfo1',
            size: '1gb',
            version: BOX_VERSION
        });
        box = appStore.waitForCloudron(box.id);
        cloudron = new Cloudron(box);
    });

    it('can activate the box', function () {
        cloudron.activate(owner);
    });

    it('can login to the box', function () {
        var token = cloudron.getOauthToken(owner);
        cloudron.setCredentials(owner.password, token);
    });

    it('can enable email', function () {
        cloudron.setEmailEnabled(true);
    });

    var location = 'test' + (Math.random() * 10000).toFixed();
    it('can install app', function () {
        var manifest = appStore.getManifest(common.TESTAPP_ID, common.TESTAPP_VERSION);
        appId = cloudron.installApp(location, manifest, { ECHO_SERVER_PORT: 5151 });
    });

    it('can populate the addons', function () {
        cloudron.populateAddons(cloudron.appFqdn(location));
    });

    it('can check the addons', function () {
        cloudron.checkAddons(cloudron.appFqdn(location), owner);
    });

    it('can check the app ratelimit', function () {
        cloudron.checkAppRateLimit(cloudron.appFqdn(location));
    });

    it('can check dnsbl', function () {
        cloudron.checkDnsbl(cloudron.appFqdn(location));
    });

    function checkExposedPort(port, done) {
        var client = new net.Socket();
        var data = [ ];
        client.connect(port, cloudron.appFqdn(location), function() {
            console.log('connected to ' + port);
        });
        client.on('data', function (chunk) { data.push(chunk); });
        client.on('close', function () {
            var response = Buffer.concat(data).toString('utf8');
            if (response === 'ECHO_SERVER_PORT=' + port) return done();
            console.log('unexpected response:', response);
            done(new Error('bad response'));
        });
        client.on('error', done);
    }

    it('can access exposed port', function (done) {
        checkExposedPort(5151, done);
    });

    it('displays app graphs', function () {
        cloudron.checkAppGraphs(appId);
    });

    it('can configure app', function () {
        location = location + 'x';
        cloudron.configureApp(appId, location, null /* altDomain */, { ECHO_SERVER_PORT: 5151 });
    });

    it('can check the addons', function () {
        cloudron.checkAddons(cloudron.appFqdn(location), owner);
    });

    it('can check dnsbl', function () {
        cloudron.checkDnsbl(cloudron.appFqdn(location));
    });

    it('can access exposed port', function (done) {
        checkExposedPort(5151, done);
    });

    it('can reboot the cloudron', function () {
        cloudron.reboot();
    });

    it('can check the addons', function () {
        cloudron.checkAddons(cloudron.appFqdn(location), owner);
    });

    it('can check dnsbl', function () {
        cloudron.checkDnsbl(cloudron.appFqdn(location));
    });

    it('can access exposed port', function (done) {
        checkExposedPort(5151, done);
    });

    // clone requires a backup to work with
    it('can backup app', function () {
        cloudron.backupApp(appId);
    });

    var cloneLocation;
    it('can clone app', function () {
        cloneLocation = location + 'clone';
        cloudron.cloneApp(appId, location + 'clone', { ECHO_SERVER_PORT: 5252 });
    });

    it('can check the addons', function () {
        cloudron.checkAddons(cloudron.appFqdn(cloneLocation), owner);
    });

    it('can access exposed port', function (done) {
        checkExposedPort(5252, done);
    });

    it('can uninstall app', function () {
        cloudron.uninstallApp(appId);
    });

    it('can delete the cloudron', function () {
        appStore.deleteCloudron(box);
    });
});

